/*    xql JDBC SQL client
    Copyright (C) 2004  Nathaniel Spurling

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package xql;

import com.Ostermiller.util.CSVPrinter;
import com.Ostermiller.util.ExcelCSVPrinter;
import com.Ostermiller.util.CSVPrint;

import java.io.*;
import java.sql.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Logger;
import java.util.logging.Level;


public class Executor {
    public static  final Logger log = Logger.getLogger(Executor.class.getName());

    public void copy(String fromPreset, String toPreset, String tableName){
        Presets.Preset ps1,ps2;
        try {
            ps1 = loadPreset(fromPreset);
            ps2 = loadPreset(toPreset);
            if(ps1==null){
                log.severe("preset="+fromPreset+" not found");
                return;
            }
            if(ps2==null){
                log.severe("preset="+toPreset+" not found");
                return;
            }
        } catch (NoLoadPresetsException e) {
            log.log(Level.SEVERE,"couldn't load presets",e);
            return;
        }
        log.fine("ps1="+ps1);
        log.fine("ps2="+ps2);
        try {
            DbAccess fromAccess = new DbAccess(ps1);
            DbAccess toAccess = new DbAccess(ps2);
            Connection c1 = fromAccess.openConnection();
            Connection c2 = toAccess.openConnection();

            try {
                Statement s1 = c1.createStatement();
                String sql = Main.sql.getArgValue();
                log.info("executing sql="+sql);
                ResultSet rst = s1.executeQuery(sql);

                List<DbColumn> columns = new ArrayList<DbColumn>();

                ResultSetMetaData rsmd = rst.getMetaData();
                int cols = rsmd.getColumnCount();
                for(int i=0;i<cols;i++){
                    int index=i+1;
                    String name = rsmd.getColumnName(index);
                    columns.add(new DbColumn(index, name, rsmd.getColumnType(index)));
                }

                List<String> columnNames = new ArrayList<String>();
                for(DbColumn dbc:columns) columnNames.add(dbc.getName());
                List<String> questionMarks = new ArrayList<String>(columnNames.size());
                for (String columnName : columnNames) questionMarks.add("?");
                String insSql = "insert into " + tableName + " ("
                        + delimitedList(columnNames)
                        + ") values ("
                        + delimitedList(questionMarks)
                        + ")";
                log.info("insert sql="+insSql);
                PreparedStatement ps = c2.prepareStatement(insSql);
                int j=0;
                while(rst.next()){
                    int i=1;
                    for(DbColumn dbc:columns){
                        try {
                            dbc.setObject(rst.getObject(i),ps);
                        } catch (SQLException e) {
                            log.warning("col="+dbc+" insSQL="+insSql);
                            throw e;
                        }
                        i++;
                    }
                    try {
                        ps.executeUpdate();
                        j++;
                    } catch (SQLException e) {
                        if(!Main.force.isSet()) throw e;
                        else{
                            log.log(Level.WARNING,"Error executing (continuing anwyay because force set)",e);
                        }
                    }
                }
                log.info(j+" rows inserted");
            } finally {
                c1.close();
                c2.close();
            }
        } catch(Exception e){
            log.log(Level.SEVERE,"error",e);
//            e.printStackTrace();
            System.exit(1);
        }
    }
    private String delimitedList(List<String> values){
        StringBuilder sb = new StringBuilder();
        for(String s:values){
            sb.append(s);
            sb.append(',');
        }
        if(sb.length()>0) sb.setLength(sb.length()-1); //eat trailing comma
        return sb.toString();
    }

    public void execute()  {
        try {
            DbAccess access = new DbAccess();

            if(Main.preset.isSet()){
                String presetName = Main.preset.getArgValue();
                Presets.Preset ps = loadPreset(presetName);
                if(ps==null) return;
                else{
                    access.configureFromPreset(ps);
                }
            }

            if(Main.jdbcdriver.isSet()) access.setDriver(Main.jdbcdriver.getArgValue());
            if(Main.jdbcurl.isSet()) access.setUrl(Main.jdbcurl.getArgValue());
            if(Main.dbusername.isSet()) access.setLogin(Main.dbusername.getArgValue());
            if(Main.dbpassword.isSet()) access.setPass(Main.dbpassword.getArgValue());
            access.emitWarnings();

            String s = executeSql(access,Main.maxrows.isSet()?Main.maxrows.getArgValue():50);
            System.out.println(s);
            if(Main.savetopreset.isSet()){
                final File pFile = Presets.getPresetFile();
                final String name = Main.savetopreset.getArgValue();
                log.info("saving preset="+name+" to file="+pFile);
                Presets p=new Presets(pFile);
                access.savePreset(p, name);
            }

        } catch(Exception e){
            log.log(Level.SEVERE,"error",e);
            System.exit(1);
        }
    }

    private Presets.Preset loadPreset(String presetName) throws NoLoadPresetsException {
        final File pFile = Presets.getPresetFile();
        Presets.Preset ps=new Presets(pFile).getPreset(presetName);
        if(ps==null) log.warning("No preset named="+presetName +" found in presetfile="+pFile);
        log.finer("loadedpreset="+ps);
        return ps;
    }


    private List getSqls() throws IOException {
        List sqls = new ArrayList();
        if(Main.scriptfile.isSet()){
            for(File f:Main.scriptfile.getArgValue()){
                LineNumberReader lnr = new LineNumberReader(new BufferedReader(new FileReader(f)));
                String line;
                while((line=lnr.readLine())!=null){
                    if (line.trim().length()>0) {
                        sqls.add(line);
                    }
                    else log.fine("ignoring empty line in file="+f);
                }
                lnr.close();
            }
        }       //sqlfile
        else if(Main.sqlfile.isSet()){
            for(File f:Main.sqlfile.getArgValue()){
                LineNumberReader lnr = new LineNumberReader(new BufferedReader(new FileReader(f)));
                StringBuffer sb = new StringBuffer();
                String line;
                while((line=lnr.readLine())!=null){
                    sb.append(line+"\n");
                }
                sqls.add(sb.toString());
                lnr.close();
            }
        }
        else{
            final String sqlString = Main.sql.getArgValue();
            if (!isNull(sqlString)) {
                sqls.add(sqlString);
            } else {
                log.warning("sql string was null or empty: ignoring");
            }
        }
        return sqls;
    }
    private boolean isNull(String driver) {
        return driver==null|| "".equals(driver.trim());
    }


    /**
     * executes SQL, prints resultset to output, returns timetaken
     * @param driverVal
     * @param url
     * @param name
     * @param pass
     * @param maxRows
     * @return
     * @throws Exception
     * @throws IOException
     * @throws ClassNotFoundException
     */
    private String executeSql(DbAccess access, int maxRows) throws Exception, IOException, ClassNotFoundException {
        Connection c = access.openConnection();

        Statement s = c.createStatement();
        try {
            StringBuffer sb = new StringBuffer();
            List sqls = getSqls();
            if(sqls.size()==0) log.warning("no sqls to execute");
            int resno=0;
            for (Iterator iterator = sqls.iterator(); iterator.hasNext();resno++) {
                String sql = (String) iterator.next();
                log.info("executing sql="+sql);
                long timetaken,time = System.currentTimeMillis();
                try {
                    if(Main.updatequery.isSet()){
                        sb.append("dbms returned: "+s.executeUpdate(sql)+" (rows affected)");
                        timetaken = System.currentTimeMillis()-time;
                    }
                    else{
                        boolean b = s.execute(sql);
                        timetaken = System.currentTimeMillis()-time;
                        boolean moreResults=false;
                        int updateCount=-1;
                        log.finer("b = " + b);
                        while(b || (moreResults=s.getMoreResults()) || (updateCount=s.getUpdateCount())!=-1){
                            log.finer("moreResults = " + moreResults);
                            log.finer("updateCount = " + updateCount);
                            if(b||moreResults){
                                ResultSet rst =s.getResultSet();
                                if (rst!=null) {
                                    if (!Main.format.isSet() || Main.format.getArgValue()== Format.text) {
                                        if(Main.transposeprintout.isSet()){
                                            printTransposedResultSet(sb, rst, maxRows);
                                        }
                                        else{
                                            printResultSet(sb, rst, maxRows);
                                        }
                                    }
                                    else{
                                         printResultSet1(sb, rst, maxRows, Main.format.getArgValue());
                                    }
                                } else {
                                    log.warning("rst was null! (ignoring)");
                                }
                            }
                            else{
                                System.out.println("updateCount = " + updateCount);
                            }
                            b=false;
                        }
                    }
                    if(Main.printtime.isSet()){
                        sb.append("took "+timetaken+" millis\n");
                    }
                    if(iterator.hasNext()){
                        sb.append("\n---------------resultset"+(resno+1)+"--------------\n");
                    }
                } catch (SQLException e) {
                    if(!Main.force.isSet()) throw e;
                    else{
                        log.log(Level.WARNING,"Error executing (continuing anwyay because force set)",e);
                    }
                }
            }
            return sb.toString();
        } finally {
            if(s!=null){
                s.close();
            }
            c.close();
        }
    }

    private void printTransposedResultSet(StringBuffer sb, ResultSet rst, int maxRows) throws SQLException {
        ResultSetMetaData rsmd = rst.getMetaData();
        boolean printRownums=(!Main.dontprintrownums.isSet());
        int row=0;
        //  boolean printSuffix;
        int columnCount = rsmd.getColumnCount();
        log.finer("printing transposed resultset printRownums="+printRownums+" maxRows="+maxRows+" colcount="+columnCount);
        while(rst.next() && (maxRows==-1 || row<maxRows)){
            //   printSuffix=row!=0;
            for(int col = 0;col< columnCount;col++){
                sb.append(rsmd.getColumnName(col+1));
                sb.append(":\t");
                sb.append(rst.getString(col+1));
                sb.append("\n");
            }
            sb.append((printRownums?String.valueOf(row):"---")+"------------\n");
            row++;
        }
    }

    private void printResultSet(StringBuffer sb, ResultSet rst, int maxRows) throws SQLException {
        ResultSetMetaData rsmd = rst.getMetaData();
        boolean printRownums=(!Main.dontprintrownums.isSet());
        for(int i=1;i<=rsmd.getColumnCount();i++){
            sb.append(rsmd.getColumnName(i));
            sb.append("("+rsmd.getColumnTypeName(i)+")");
            if(i<rsmd.getColumnCount()){
                sb.append("\t");
            }
        }
        sb.append("\n------------------\n");
        //   boolean printSuffix;
        int row=0;
        while(rst.next() && (maxRows==-1 || row<maxRows)){
            // printSuffix=row!=0;
            for(int col = 0;col<rsmd.getColumnCount();col++){
                sb.append(rst.getString(col+1));
                if(col+1<rsmd.getColumnCount()){
                    sb.append("\t");
                }
            }
            //   if(printSuffix){
            int realRow = row+1;
            if(printRownums && realRow!=0 && realRow%10==0){
                sb.append("\n------"+realRow+"-----\n");
            }
            else sb.append("\n");
            //  }
            row++;
        }
    }

    private void printResultSet1(StringBuffer sb, ResultSet rst, int maxRows, Format format) throws SQLException, IOException {
        ByteArrayOutputStream os = new ByteArrayOutputStream();
        final CSVPrint csvp;
        if(format== Format.excel_csv){
             csvp=new ExcelCSVPrinter(os);
        }
        else if(format== Format.csv){
            csvp=new CSVPrinter(os);
        }
        else throw new RuntimeException("unrecognized format="+format);

        ResultSetMetaData rsmd = rst.getMetaData();
        String line[] = new String[rsmd.getColumnCount()];
        for(int i=1;i<=rsmd.getColumnCount();i++){
            line[i-1]=rsmd.getColumnName(i);
        }
        csvp.writeln(line);
        //   boolean printSuffix;
        int row=0;
        while(rst.next() && (maxRows==-1 || row<maxRows)){
            for(int col = 0;col<rsmd.getColumnCount();col++){
                line[col]=rst.getString(col+1);
            }
            csvp.writeln(line);
            row++;
        }
        sb.append(os.toString("UTF8"));
    }

}
